/************************************************************************************
RMP RT-Component
Copyright (c) 2010, Segway Japan, Ltd.
All rights reserved.

Contact us if you use this software for sell.
If you use this software not for sell, you can use this software under BSD lisence.
See the files LICENSE.TXT and LICENSE-BSD.TXT for more details.                     
************************************************************************************/
#include <math.h>
#include "RMP200Controller.h"


#define USE_SLIP  (1)


// Module specification
// <rtc-template block="module_spec">
static const char* joystickcontroller_spec[] =
  {
    "implementation_id", "VirtualRMP200Controller",
    "type_name",         "VirtualRMP200Controller",
    "description",       "Controller to drive a mobile robot",
    "version",           "1.0.0",
    "vendor",            "AIST HRG",
    "category",          "OpenHRP Controller",
    "activity_type",     "DataFlowComponent",
    "max_instance",      "1",
    "language",          "C++",
    "lang_type",         "compile",
    // Configuration variables
	"conf.default.P", "5.0",
    "conf.default.I", "100.0",
    "conf.default.D", "1.0",
    "conf.default.KPID", "1.0",

	"conf.default.Kspeed", "1.0",
    "conf.default.Kturn", "1.0",
    "conf.default.Kodoxy", "1.0",
    "conf.default.Kodotheta", "1.0",
    "conf.default.Kslip", "0.0",
    "conf.default.mps2rmp", "160.0",
    "conf.default.ODOoutput", "0",
    "conf.default.idIIS", "200",
    "conf.default.wheelwidth", "0.54",
    "conf.default.wheelradiusL", "0.235",
    "conf.default.wheelradiusR", "0.235",
    "conf.default.wheelcount1rot", "49646.0775",
    "conf.default.MaxV", "2.0",
    "conf.default.MaxW", "2.0",


    ""
  };
// </rtc-template>

RMP200Controller::RMP200Controller(RTC::Manager* manager)
  : RTC::DataFlowComponentBase(manager),
    // <rtc-template block="initializer">
    m_angleIn("angle", m_angle),
    m_velocityIn("velocity", m_velocity),
    m_commandIn("command", m_command),
    m_InputIn("Input", m_Input),
    m_VelocityIn("Velocity", m_Velocity),
    m_VelocityIISIn("VelocityIIS", m_VelocityIIS),
    m_Velocity2DIISIn("Velocity2DIIS", m_Velocity2DIIS),

	m_torqueOut("torque", m_torque),
	m_EncoderLOut("EncoderL", m_EncoderL),
    m_EncoderROut("EncoderR", m_EncoderR),
    m_OdometryOut("Odometry", m_Odometry),
    m_OdometryBOut("OdometryB", m_OdometryB),
    m_RobotVelIISOut("RobotVelIIS", m_RobotVelIIS),
    m_RobotPosIISOut("RobotPosIIS", m_RobotPosIIS),    
    m_RobotVel2DIISOut("RobotVel2DIIS", m_RobotVel2DIIS),
    m_RobotPose2DIISOut("RobotPose2DIIS", m_RobotPose2DIIS),

    // </rtc-template>
	dummy(0)
{
  // Registration: InPort/OutPort/Service
  // <rtc-template block="registration">
  // Set InPort buffers
  registerInPort("angle", m_angleIn);
  registerInPort("velocity", m_velocityIn);
  registerInPort("command", m_commandIn);

  registerInPort("Input", m_InputIn);
  registerInPort("Velocity", m_VelocityIn);
  registerInPort("VelocityIIS", m_VelocityIISIn);
  registerInPort("Velocity2DIIS", m_Velocity2DIISIn);
  
  // Set OutPort buffer
  registerOutPort("torque", m_torqueOut);

  registerOutPort("EncoderL", m_EncoderLOut);
  registerOutPort("EncoderR", m_EncoderROut);
  registerOutPort("Odometry", m_OdometryOut);
  registerOutPort("OdometryB", m_OdometryBOut);
  registerOutPort("RobotVelIIS", m_RobotVelIISOut);
  registerOutPort("RobotPosIIS", m_RobotPosIISOut);
  registerOutPort("RobotVel2DIIS", m_RobotVel2DIISOut);
  registerOutPort("RobotPose2DIIS", m_RobotPose2DIISOut);
  
  // Set service provider to Ports
  
  // Set service consumers to Ports
  
  // Set CORBA Service Ports
  
  // </rtc-template>

}

RMP200Controller::~RMP200Controller()
{
}


RTC::ReturnCode_t RMP200Controller::onInitialize()
{
  // <rtc-template block="bind_config">
  // Bind variables and configuration variable
  bindParameter("P", m_P, "1.0");
  bindParameter("I", m_I, "1.0");
  bindParameter("D", m_D, "1.0");
  bindParameter("KPID", m_KPID, "1.0");

  bindParameter("Kspeed", m_Kspeed, "1.0");
  bindParameter("Kturn", m_Kturn, "1.0");
  bindParameter("Kodoxy", m_Kodoxy, "1.0");
  bindParameter("Kodotheta", m_Kodotheta, "1.0");
  bindParameter("Kslip", m_Kslip, "0.0");
  bindParameter("mps2rmp", m_mps2rmp, "160.0");
  bindParameter("ODOoutput", m_ODOoutput, "0");
  bindParameter("idIIS", m_idIIS, "200");
  bindParameter("wheelwidth", m_wheelwidth, "0.54");
  bindParameter("wheelradiusL", m_wheelradiusL, "0.235");
  bindParameter("wheelradiusR", m_wheelradiusR, "0.235");
  bindParameter("wheelcount1rot", m_wheelcount1rot, "49646.0775");
  bindParameter("MaxV", m_MaxV, "2.0");
  bindParameter("MaxW", m_MaxW, "2.0");

  // </rtc-template>

  // |[g
  m_torque.data.length(4);

  tireVelFL = 0.0;
  tireVelFR = 0.0;
  tireVelRL = 0.0;
  tireVelRR = 0.0;
  tireVelFLint = 0.0;
  tireVelFRint = 0.0;
  tireVelRLint = 0.0;
  tireVelRRint = 0.0;

  Speed = 0;
  Turn = 0;

  return RTC::RTC_OK;
}

/*
RTC::ReturnCode_t RMP200Controller::onFinalize()
{
  return RTC::RTC_OK;
}
*/

/*
RTC::ReturnCode_t RMP200Controller::onStartup(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/

/*
RTC::ReturnCode_t RMP200Controller::onShutdown(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/

RTC::ReturnCode_t RMP200Controller::onActivated(RTC::UniqueId ec_id)
{
  Status.wheelwidth = m_wheelwidth;
  Status.wheelradiusL = m_wheelradiusL;
  Status.wheelradiusR = m_wheelradiusR;
  Status.count1rot = m_wheelcount1rot;

  Stop = 0;

  Status.odotime = 
  Status.odotime_prev = coil::gettimeofday();

  curtime = coil::gettimeofday();

  return RTC::RTC_OK;
}

/*
RTC::ReturnCode_t RMP200Controller::onDeactivated(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/

//Torque output for virtual robot
void RMP200Controller::rmp_set_speed(int l, int r)
{

  double bs2virtual = 0.2;//ratio of Blackship Speed to Virtual Robot Speed

  double tireCommandVelL = (double)l * bs2virtual;
  double tireCommandVelR = (double)r * bs2virtual;

  double tireVelFLold = tireVelFL;
  double tireVelFRold = tireVelFR;

  m_velocityIn.read();
  tireVelFL = m_velocity.data[0];
  tireVelFR = m_velocity.data[1];
  //tireVelRL = m_velocity.data[2];
  //tireVelRR = m_velocity.data[3];

  double tireVelFLdiff = tireCommandVelL - tireVelFL;
  double tireVelFRdiff = tireCommandVelR - tireVelFR;
  //double tireVelRLdiff = tireCommandVelL - tireVelRL;
  //double tireVelRRdiff = tireCommandVelR - tireVelRR;


  // 쓮gNvZ(PID)
  double P = m_P;
  double I = m_I;
  double D = m_D;
  double K = m_KPID;
  double dt = 1.0;
  P *= K;
  I *= K;
  D *= K;
  //P
  double tireCommandTorqueFL = P * tireVelFLdiff;
  double tireCommandTorqueFR = P * tireVelFRdiff;
  //double tireCommandTorqueRL = P * tireVelRLdiff;
  //double tireCommandTorqueRR = P * tireVelRRdiff;
  //I
  tireVelFLint += tireVelFLdiff * dt;
  tireVelFRint += tireVelFRdiff * dt;
  //tireVelRLint += tireVelRLdiff * dt;
  //tireVelRRint += tireVelRRdiff * dt;
  tireCommandTorqueFL += I * tireVelFLint;
  tireCommandTorqueFR += I * tireVelFRint;
  //tireCommandTorqueRL += I * tireVelRLint;
  //tireCommandTorqueRR += I * tireVelRRint;
  //D
  tireCommandTorqueFL -= D * (tireVelFL - tireVelFLold) / dt;
  tireCommandTorqueFR -= D * (tireVelFR - tireVelFRold) / dt;
  //tireCommandTorqueRL -= D * (tireVelRL - tireVelRLold) / dt;
  //tireCommandTorqueRR -= D * (tireVelRR - tireVelRRold) / dt;

  m_torque.data[0] = tireCommandTorqueFL;
  m_torque.data[1] = tireCommandTorqueFR;
  m_torque.data[2] = 0.0;
  m_torque.data[3] = 0.0;
  m_torqueOut.write(); 
}

//Encoder input from virtual robot
void RMP200Controller::rmp_get_encoder2(double count1rot, int &left, int &right)
{
  m_angleIn.read();
  double wheelangle[4];
  wheelangle[0] = m_angle.data[0];
  wheelangle[1] = m_angle.data[1];
  //wheelangle[2] = m_angle.data[2];
  //wheelangle[3] = m_angle.data[3];

  double langle = wheelangle[0];
  double rangle = wheelangle[1];

  double pi = 3.1415926535897932384626433832795;
#if 0
  int lcount;
  lcount = (int) (count1rot * langle / (2.0*pi));
  lcount %= 65536;
  if (lcount < 0) {
	  lcount = 65536 + lcount;
  }
  int rcount;
  rcount = (int) (count1rot * rangle / (2.0*pi));
  rcount %= 65536;
  if (rcount < 0) {
	  rcount = 65536 + rcount;
  }
#else
  left = (int) (count1rot * langle / (2.0*pi));
  right = (int) (count1rot * rangle / (2.0*pi));
#endif
}

void rmp_odometry(rmp_status& status, double kxy, double ktheta, double kslip, bool ODOoutput)
{
  double pi = 3.1415926535;

  //setting
  double wheelrL = status.wheelradiusL;
  double wheelrR = status.wheelradiusR;
  double radius = status.wheelwidth * 0.5;
  double count1rot = status.count1rot;

  if (status.odofirst) {
    status.int_left_prev  = status.int_left_wheel;
    status.int_right_prev = status.int_right_wheel;
    status.odofirst = false;
  }


  const static double meterpercountL = 2.0 * wheelrL * pi / count1rot;
  const static double meterpercountR = 2.0 * wheelrR * pi / count1rot;
  const static int iterate = 64;

  double countL = (double)(status.int_left_wheel  - status.int_left_prev);
  double countR = (double)(status.int_right_wheel - status.int_right_prev);


  countR /= (double)iterate;
  countL /= (double)iterate;

  double meterR = countR * meterpercountR;
  double meterL = countL * meterpercountL;

#if USE_SLIP
  double diff = meterR - meterL;
  double slip = diff * kslip;
  double meterRslip = meterR - slip;
  double meterLslip = meterL + slip;
  double forward = (meterRslip + meterLslip) * 0.5;
  double diffR = meterRslip - forward;
#else
  double forward = (meterR + meterL) * 0.5;
  double diffR = meterR - forward;
#endif
  double addtheta = diffR / radius;


  status.addx = 0.0;
  status.addy = 0.0;
  status.addtheta = 0.0;
  for (int i = 0; i < iterate; i++) {
    status.odoy += forward * cos(status.odotheta) * kxy;
    status.odox -= forward * sin(status.odotheta) * kxy;
    status.odotheta += addtheta * ktheta;

    status.addy += forward * kxy;
    status.addx += 0.0;
    status.addtheta += addtheta * ktheta;
  }


  if (ODOoutput) {
    printf("odo %f %f %f enc %d %d   wheelrL %f wheelrR %f radius %f count1rot %f\n",
	   (float)status.odox, (float)status.odoy, (float)status.odotheta,
	   status.int_left_wheel, status.int_right_wheel,
	   wheelrL, wheelrR, radius, count1rot);
  }


  status.int_left_prev  = status.int_left_wheel;
  status.int_right_prev = status.int_right_wheel;
}


//Blackship.cpp BlackshipCompƓ
RTC::ReturnCode_t RMP200Controller::onExecute(RTC::UniqueId ec_id)
{

  if (m_InputIn.isNew()) {
    m_InputIn.read();
    Stop = m_Input.data;
  }

  double mps2rmp = m_mps2rmp;

    if (m_VelocityIn.isNew() || 
      m_VelocityIISIn.isNew() || 
      m_Velocity2DIISIn.isNew()) {

    double speedmps;
    double turnrps;

    if (m_VelocityIn.isNew()) {
      while (!m_VelocityIn.isEmpty()) {
	m_VelocityIn.read();
      }
      speedmps = m_Velocity.v;
      turnrps = m_Velocity.w;
    }
    if (m_VelocityIISIn.isNew()) {
      while (!m_VelocityIISIn.isEmpty()) {
	m_VelocityIISIn.read();
      }
      speedmps = m_VelocityIIS.vx;
      turnrps = m_VelocityIIS.w;
    }
    if (m_Velocity2DIISIn.isNew()) {
      while (!m_Velocity2DIISIn.isEmpty()) {
	m_Velocity2DIISIn.read();
      }

      speedmps = m_Velocity2DIIS.data.vx;
      turnrps = m_Velocity2DIIS.data.va;
    }

    if (fabs(speedmps) > m_MaxV) {
      speedmps = (speedmps > 0.0)? m_MaxV : -m_MaxV;
    }
    if (fabs(turnrps) > m_MaxW) {
      turnrps = (turnrps > 0.0)? m_MaxW : -m_MaxW;
    }
 
    Speed = (int) (mps2rmp * m_Kspeed * speedmps);
#if USE_SLIP
    Turn = (int) (mps2rmp * m_Kturn * (m_wheelwidth*0.5 * turnrps) / (1.0 - 2.0 * m_Kslip));
#else
    Turn = (int) (mps2rmp * m_Kturn * m_wheelwidth*0.5 * turnrps);
#endif
  }



  if (Stop) {
    Speed = 0;
    Turn = 0;
  }

  int s[2];
  s[0] = Speed;
  s[1] = Speed;
  s[0] -= Turn;
  s[1] += Turn;

#if 0
  rmp_set_speed(PortFd, s[0], s[1]);
#else
  rmp_set_speed(s[0], s[1]);
#endif


  //-------------------------------
  //get encoder
  //-------------------------------
  int encoder[2];
  Status.odotime_prev = Status.odotime;
#if 0
  blackship_get_encoder2(PortFd, encoder[0], encoder[1]);
#else
  rmp_get_encoder2(Status.count1rot, encoder[0], encoder[1]);
#endif
  Status.odotime = coil::gettimeofday();

  Status.int_left_wheel = encoder[0];
  Status.int_right_wheel = encoder[1];
  rmp_odometry(Status, m_Kodoxy, m_Kodotheta, m_Kslip, m_ODOoutput);


  //--------------------------------------------------------------------------------
  //output velocity
  //  two wheeled blackship can not run at 1 m/sec. max speed is about 0.4m/sec
  //
  //--------------------------------------------------------------------------------
#if 0
  if (m_simulator == 0 ||
      fabs((float)Status.addx) > 0.0 ||
      fabs((float)Status.addy) > 0.0 || 
      fabs((float)Status.addtheta) > 0.0) {
#endif
   //culc speed
    coil::TimeValue tdiff = Status.odotime - Status.odotime_prev;
    double sec    = (double)((double)tdiff.sec() + (double)tdiff.usec() / 1000000.0);
    double dx     = Status.addx / sec;     //(m/sec)
    double dy     = Status.addy / sec;     //(m/sec)
    double dtheta = Status.addtheta / sec; //(rad/sec)

    //SJ to IIS coord
    double dxIIS = dy;
    double dyIIS = -dx;
    double dthetaIIS = dtheta;

    m_RobotVelIIS.id.length(1);
    m_RobotVelIIS.id[0]   = m_idIIS;
    m_RobotVelIIS.tm.sec  = (::CORBA::ULong) Status.odotime.sec();
    m_RobotVelIIS.tm.nsec = (::CORBA::ULong) (Status.odotime.usec() * 1000);
    m_RobotVelIIS.vx      = dxIIS;
    m_RobotVelIIS.vy      = dyIIS;
    m_RobotVelIIS.w       = dthetaIIS;

    m_RobotVel2DIIS.id.length(1);
    m_RobotVel2DIIS.id[0]   = m_idIIS;
    m_RobotVel2DIIS.tm.sec  = (::CORBA::ULong) Status.odotime.sec();
    m_RobotVel2DIIS.tm.nsec = (::CORBA::ULong) (Status.odotime.usec() * 1000);
    m_RobotVel2DIIS.data.vx = dxIIS;
    m_RobotVel2DIIS.data.vy = dyIIS;
    m_RobotVel2DIIS.data.va = dthetaIIS;
#if 0    
  }
#endif

  {
    //SJ to IIS coord
    m_RobotPosIIS.id.length(1);
    m_RobotPosIIS.id[0]   = m_idIIS;
    m_RobotPosIIS.tm.sec  = (::CORBA::ULong) Status.odotime.sec();
    m_RobotPosIIS.tm.nsec = (::CORBA::ULong) (Status.odotime.usec() * 1000);
    m_RobotPosIIS.x       = Status.odoy;
    m_RobotPosIIS.y       = -Status.odox;
    m_RobotPosIIS.theta   = Status.odotheta;

    m_RobotPose2DIIS.id.length(1);
    m_RobotPose2DIIS.id[0]   = m_idIIS;
    m_RobotPose2DIIS.tm.sec  = (::CORBA::ULong) Status.odotime.sec();
    m_RobotPose2DIIS.tm.nsec = (::CORBA::ULong) (Status.odotime.usec() * 1000);
    m_RobotPose2DIIS.data.position.x       = Status.odoy;
    m_RobotPose2DIIS.data.position.y       = -Status.odox;
    m_RobotPose2DIIS.data.heading          = Status.odotheta;

    //outodo
    //printf("odo %f %f %f\n", m_RobotPosIIS.x, m_RobotPosIIS.y, m_RobotPosIIS.theta);
  }


  

  RTC::Time tmodo;
  tmodo.sec  = (::CORBA::ULong) Status.odotime.sec();
  tmodo.nsec = (::CORBA::ULong) (Status.odotime.usec() * 1000);

  m_EncoderL.tm     = tmodo;
  m_EncoderL.data   = encoder[0];
  m_EncoderR.tm     = tmodo;
  m_EncoderR.data   = encoder[1];

  m_Odometry.tm     = tmodo;
  m_Odometry.x      = Status.odox;
  m_Odometry.y      = Status.odoy;
  m_Odometry.theta  = Status.odotheta;

  m_OdometryB.tm    = tmodo;
  m_OdometryB.x     = m_RobotPosIIS.x;
  m_OdometryB.y     = m_RobotPosIIS.y;
  m_OdometryB.theta = m_RobotPosIIS.theta;


  m_EncoderLOut.write();
  m_EncoderROut.write();
  m_OdometryOut.write();
  m_OdometryBOut.write();
  

  //IIS
  m_RobotVelIISOut.write();
  m_RobotVel2DIISOut.write();
  
  m_RobotPosIISOut.write();
  m_RobotPose2DIISOut.write();



  return RTC::RTC_OK;
}

/*
RTC::ReturnCode_t RMP200Controller::onAborting(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/

/*
RTC::ReturnCode_t RMP200Controller::onError(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/

/*
RTC::ReturnCode_t RMP200Controller::onReset(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/

/*
RTC::ReturnCode_t RMP200Controller::onStateUpdate(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/

/*
RTC::ReturnCode_t RMP200Controller::onRateChanged(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/



extern "C"
{
 
  void RMP200ControllerInit(RTC::Manager* manager)
  {
    RTC::Properties profile(joystickcontroller_spec);
    manager->registerFactory(profile,
                             RTC::Create<RMP200Controller>,
                             RTC::Delete<RMP200Controller>);
  }
  
};


