/************************************************************************************
RMP RT-Component
Copyright (c) 2010, Segway Japan, Ltd.
All rights reserved.

Contact us if you use this software for sell.
If you use this software not for sell, you can use this software under BSD lisence.
See the files LICENSE.TXT and LICENSE-BSD.TXT for more details.                     
************************************************************************************/
#ifndef RMP_H
#define RMP_H

#include <rtm/Manager.h>
#include <rtm/DataFlowComponentBase.h>
#include <rtm/CorbaPort.h>
#include <rtm/DataInPort.h>
#include <rtm/DataOutPort.h>
#include <rtm/idl/BasicDataTypeSkel.h>

#include "rmp_if.h"
#include "intellirobotStub.h"

// Service implementation headers
// <rtc-template block="service_impl_h">

// </rtc-template>

// Service Consumer stub headers
// <rtc-template block="consumer_stub_h">

// </rtc-template>
using namespace RTC;

/*!
  @brief RMP I-O RTM-Module class\n
   inport TimedLong 'Speed' [-1176 - 1176] is sended to RMP directly\n
   inport TimedLong 'Turn'  [-1024 - 1024] is sended to RMP directly\n
   inport TimedDouble 'Speed_mps' (m/s)\n
   inport TimedDouble 'Turn_rps'  (rad/s)\n
   inport TimedVelocity (m/s rad/s)
   outport TimedLong 'EndoerL','EncoderR' are recved from RMP directly\
   outport TimedFloat 'Battery','UiBattery' are voltage.\n
   outport TimedFloat 'PitchAngle' is angle (not radian).\n
   outport TimedOdometry 'Odometry' is internal odometry x(m) y(m) theta(rad) \n
\n
   configuration
\n
   ===real RMP===\n
   str_port:std::string:FTDI\n
\n
   ===simulated RMP===\n
   str_port:std::string:/dev/ttySIM*

*/
class Rmp
  : public RTC::DataFlowComponentBase
{
 public:
  Rmp(RTC::Manager* manager);
  ~Rmp();

  // The initialize action (on CREATED->ALIVE transition)
  // formaer rtc_init_entry()
 virtual RTC::ReturnCode_t onInitialize();

  // The finalize action (on ALIVE->END transition)
  // formaer rtc_exiting_entry()
  // virtual RTC::ReturnCode_t onFinalize();

  // The startup action when ExecutionContext startup
  // former rtc_starting_entry()
  // virtual RTC::ReturnCode_t onStartup(RTC::UniqueId ec_id);

  // The shutdown action when ExecutionContext stop
  // former rtc_stopping_entry()
  // virtual RTC::ReturnCode_t onShutdown(RTC::UniqueId ec_id);

  // The activated action (Active state entry action)
  // former rtc_active_entry()
  virtual RTC::ReturnCode_t onActivated(RTC::UniqueId ec_id);

  // The deactivated action (Active state exit action)
  // former rtc_active_exit()
  virtual RTC::ReturnCode_t onDeactivated(RTC::UniqueId ec_id);

  // The execution action that is invoked periodically
  // former rtc_active_do()
  virtual RTC::ReturnCode_t onExecute(RTC::UniqueId ec_id);

  // The aborting action when main logic error occurred.
  // former rtc_aborting_entry()
  // virtual RTC::ReturnCode_t onAborting(RTC::UniqueId ec_id);

  // The error action in ERROR state
  // former rtc_error_do()
  // virtual RTC::ReturnCode_t onError(RTC::UniqueId ec_id);

  // The reset action that is invoked resetting
  // This is same but different the former rtc_init_entry()
  // virtual RTC::ReturnCode_t onReset(RTC::UniqueId ec_id);

  // The state update action that is invoked after onExecute() action
  // no corresponding operation exists in OpenRTm-aist-0.2.0
  // virtual RTC::ReturnCode_t onStateUpdate(RTC::UniqueId ec_id);

  // The action that is invoked when execution context's rate is changed
  // no corresponding operation exists in OpenRTm-aist-0.2.0
  // virtual RTC::ReturnCode_t onRateChanged(RTC::UniqueId ec_id);


 protected:
  // Configuration variable declaration
  // <rtc-template block="config_declare">
  /*!
   * 
   * - Name:  Kspeed
   * - DefaultValue: 1.0
   */
  double m_Kspeed;
  /*!
   * 
   * - Name:  KspeedA
   * - DefaultValue: -0.01630197
   */
  double m_KspeedA;
  /*!
   * 
   * - Name:  KspeedB
   * - DefaultValue: 1.412417763
   */
  double m_KspeedB;
  /*!
   * 
   * - Name:  Kturn
   * - DefaultValue: 1.0
   */
  double m_Kturn;
  /*!
   * 
   * - Name:  wheelwidth
   * - DefaultValue: 0.54
   */
  double m_wheelwidth;
  /*!
   * 
   * - Name:  wheelradiusL
   * - DefaultValue: 0.235
   */
  double m_wheelradiusL;
  /*!
   * 
   * - Name:  wheelradiusR
   * - DefaultValue: 0.235
   */
  double m_wheelradiusR;
  /*!
   * 
   * - Name:  wheelcount1rot
   * - DefaultValue: 49646.0775
   */
  double m_wheelcount1rot;
  /*!
   * 
   * - Name:  str_port
   * - DefaultValue: FTDI
   */
  std::string m_str_port;
  /*!
   * 
   * - Name:  Kodoxy
   * - DefaultValue: 1.0
   */
  double m_Kodoxy;
  /*!
   * 
   * - Name:  Kodotheta
   * - DefaultValue: 1.0
   */
  double m_Kodotheta;
  /*!
   * 
   * - Name:  simulator
   * - DefaultValue: 0
   */
  int m_simulator;
  /*!
   * 
   * - Name:  startxIIS
   * - DefaultValue: 0.0
   */
  int m_idIIS;

  double m_startxIIS;
  /*!
   * 
   * - Name:  startyIIS
   * - DefaultValue: 0.0
   */
  double m_startyIIS;
  /*!
   * 
   * - Name:  startthetaIIS
   * - DefaultValue: 0.0
   */
  double m_startthetaIIS;
  /*!
   * 
   * - Name:  Mode
   * - DefaultValue: 0
   */
  int m_Mode;
  /*!
   * 
   * - Name:  MaxV
   * - DefaultValue: 2.0
   */
  double m_MaxV;
  /*!
   * 
   * - Name:  MaxW
   * - DefaultValue: 2.0
   */
  double m_MaxW;
  // </rtc-template>

  // DataInPort declaration
  // <rtc-template block="inport_declare">
  RTC::TimedULong m_Input;
  /*!
   */
  InPort<RTC::TimedULong> m_InputIn;
  RTC::TimedVelocity m_Velocity;
  /*!
   */
  InPort<RTC::TimedVelocity> m_VelocityIn;
  IIS::TimedVelocity m_VelocityIIS;
  /*!
   */
  InPort<IIS::TimedVelocity> m_VelocityIISIn;
  IIS::TimedVelocity2D m_Velocity2DIIS;
  /*!
   */
  InPort<IIS::TimedVelocity2D> m_Velocity2DIISIn;
  
  // </rtc-template>

  // DataOutPort declaration
  // <rtc-template block="outport_declare">
  RTC::TimedLong m_EncoderL;
  /*!
   */
  OutPort<RTC::TimedLong> m_EncoderLOut;
  RTC::TimedLong m_EncoderR;
  /*!
   */
  OutPort<RTC::TimedLong> m_EncoderROut;
  RTC::TimedFloat m_Battery;
  /*!
   */
  OutPort<RTC::TimedFloat> m_BatteryOut;
  RTC::TimedFloat m_UiBattery;
  /*!
   */
  OutPort<RTC::TimedFloat> m_UiBatteryOut;
  RTC::TimedFloat m_PitchAngle;
  /*!
   */
  OutPort<RTC::TimedFloat> m_PitchAngleOut;
  RTC::TimedFloat m_PitchRate;
  /*!
   */
  OutPort<RTC::TimedFloat> m_PitchRateOut;
  RTC::TimedFloat m_RollAngle;
  /*!
   */
  OutPort<RTC::TimedFloat> m_RollAngleOut;
  RTC::TimedFloat m_RollRate;
  /*!
   */
  OutPort<RTC::TimedFloat> m_RollRateOut;
  RTC::TimedFloat m_YawRate;
  /*!
   */
  OutPort<RTC::TimedFloat> m_YawRateOut;
  RTC::TimedOdometry m_Odometry;
  /*!
   */
  OutPort<RTC::TimedOdometry> m_OdometryOut;

  RTC::TimedOdometry m_OdometryB;
  /*!
   */
  OutPort<RTC::TimedOdometry> m_OdometryBOut;


  IIS::TimedVelocity m_RobotVelIIS;
  /*!
   */
  OutPort<IIS::TimedVelocity> m_RobotVelIISOut;
  IIS::TimedPosition m_RobotPosIIS;
  /*!
   */
  OutPort<IIS::TimedPosition> m_RobotPosIISOut;
  IIS::TimedVelocity2D m_RobotVel2DIIS;
  /*!
   */
  OutPort<IIS::TimedVelocity2D> m_RobotVel2DIISOut;
  IIS::TimedPose2D m_RobotPose2DIIS;
  /*!
   */
  OutPort<IIS::TimedPose2D> m_RobotPose2DIISOut;


  
  // </rtc-template>

  // CORBA Port declaration
  // <rtc-template block="corbaport_declare">
  
  // </rtc-template>

  // Service declaration
  // <rtc-template block="service_declare">
  
  // </rtc-template>

  // Consumer declaration
  // <rtc-template block="consumer_declare">
  
  // </rtc-template>

 private:
  int dummy;

  int Speed;
  int Turn;

  rmp_status Status;

  unsigned long Stop;
};


extern "C"
{
  void RmpInit(RTC::Manager* manager);
};

#endif // RMP_H
